@@ -1,4 +1,4 @@
-# Copyright 2022 The Flax Authors.
+# Copyright 2024 The Flax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,80 +14,86 @@
 
 """Flax functional core: Scopes."""
 
+import collections
 import contextlib
 import dataclasses
 import functools
 import hashlib
 import typing
-from typing import (Any, Callable, Dict, Generic, Iterable, Mapping, Optional,
-                    Sequence, Set, Tuple, TypeVar, Union)
+from typing import (
+  Any,
+  Callable,
+  Dict,
+  Generic,
+  Iterable,
+  Literal,
+  Mapping,
+  Optional,
+  Sequence,
+  Set,
+  Tuple,
+  TypeVar,
+  Union,
+  cast,
+  overload,
+)
+
+import jax
+import numpy as np
+from jax import numpy as jnp
+from jax import random, tree_util
 
-from flax.ids import uuid
 from flax import config as config
 from flax import configurations as legacy_config  # only for flax_lazy_rng
-from flax import errors
-from flax import struct
-from flax import traceback_util
-from .frozen_dict import freeze
-from .frozen_dict import FrozenDict
-from .frozen_dict import unfreeze
-from . import partial_eval
-from . import tracers
-from . import meta
-import jax
-from jax import config as jax_config
-from jax import numpy as jnp
-from jax import random
-from jax import tree_util
-import numpy as np
+from flax import errors, struct, traceback_util
+from flax.ids import uuid
+from flax.typing import (
+  PRNGKey,
+  Array,
+  RNGSequences,
+  Collection,
+  MutableCollection,
+  VariableDict,
+  FrozenVariableDict as FrozenVariableDict,
+  MutableVariableDict,
+  PRNGFoldable
+)
+
+from . import meta, partial_eval, tracers
+from .frozen_dict import FrozenDict, freeze, unfreeze
 
 traceback_util.register_exclusion(__file__)
 
 T = TypeVar('T')
 
-PRNGKey = Any
-Array = Any
-
-RNGSequences = Dict[str, PRNGKey]
-
 
 Filter = Union[bool, str, typing.Collection[str], 'DenyList']
 
 # When conditioning on filters we require explicit boolean comparisons.
 # pylint: disable=g-bool-id-comparison
 
+
 @dataclasses.dataclass(frozen=True, eq=True)
 class DenyList:
   """DenyList represents an opt-out based mutability filter.
-
   DenyList can be used to make every collection mutable except the ones
   defined in the given filter.
   To for example make everything but the params collection mutable::
-
     nn.apply(fn, mutable=nn.DenyList(["params"]))
-
   Attributes:
     deny: The filter representing the collections that are not mutable.
-
   """
+
   deny: Filter
 
 
 CollectionFilter = Filter
 PRNGSequenceFilter = Filter
 
-Collection = Mapping[str, Any]
-MutableCollection = Dict[str, Any]
-
-VariableDict = Mapping[str, Collection]
-FrozenVariableDict = FrozenDict[str, Collection]
-MutableVariableDict = Dict[str, MutableCollection]
-
-PRNGFoldable = Union[int, str]
-
 
 class LazyRng(struct.PyTreeNode):
   """Wrapper around JAX PRNGKey that lazily maintains a tuple of static data to be folded into the rng."""
+
   rng: PRNGKey
   suffix: Tuple[PRNGFoldable, ...] = struct.field(pytree_node=False)
 
@@ -95,8 +101,9 @@ class LazyRng(struct.PyTreeNode):
     return _fold_in_static(self.rng, self.suffix)
 
   @staticmethod
-  def create(rng: Union['LazyRng', PRNGKey],
-             *suffix: PRNGFoldable) -> 'LazyRng':
+  def create(
+    rng: Union['LazyRng', PRNGKey], *suffix: PRNGFoldable
+  ) -> 'LazyRng':
     if not legacy_config.flax_lazy_rng:
       if isinstance(rng, LazyRng):
         assert not rng.suffix
@@ -124,8 +131,9 @@ def _legacy_rng_fold_in(rng: PRNGKey, data: Iterable[PRNGFoldable]) -> PRNGKey:
   return rng
 
 
-def _fold_in_static(rng: PRNGKey,
-                    data: typing.Collection[PRNGFoldable]) -> PRNGKey:
+def _fold_in_static(
+  rng: PRNGKey, data: typing.Collection[PRNGFoldable]
+) -> PRNGKey:
   """Folds static data (strings & ints) into a jax.random.PRNGKey using its SHA-1 hash.
 
   This is faster than splitting an PRNGKey because it allows generating new PRNG
@@ -142,6 +150,9 @@ def _fold_in_static(rng: PRNGKey,
     return rng
   m = hashlib.sha1()
   for x in data:
+    if config.flax_fix_rng_separator:
+      # encode seperate to avoid collisions like for example: ("ab", "c") and ("a", "bc")
+      m.update(b'\00')
     if isinstance(x, str):
       m.update(x.encode('utf-8'))
     elif isinstance(x, int):
@@ -150,7 +161,7 @@ def _fold_in_static(rng: PRNGKey,
       raise ValueError(f'Expected int or string, got: {x}')
   d = m.digest()
   hash_int = int.from_bytes(d[:4], byteorder='big')
-  return random.fold_in(rng, jnp.uint32(hash_int)) # type: ignore
+  return random.fold_in(rng, jnp.uint32(hash_int))  # type: ignore
 
 
 def is_filter_empty(filter_like: Filter) -> bool:
@@ -299,8 +310,8 @@ def intersect_filters(a: Filter, b: Filter) -> Filter:
 
 
 def group_collections(
-    xs: VariableDict,
-    col_filters: Sequence[CollectionFilter]) -> Sequence[MutableVariableDict]:
+  xs: VariableDict, col_filters: Sequence[CollectionFilter]
+) -> Sequence[MutableVariableDict]:
   """Groups variables by collection filters.
 
   Iteratively applies the filters in `col_filters` to `xs`, and adds the result
@@ -339,8 +350,7 @@ class Variable(Generic[T]):
   content and can be assigned to for mutation.
   """
 
-  def __init__(
-      self, scope: 'Scope', collection: str, name: str, unbox: bool):
+  def __init__(self, scope: 'Scope', collection: str, name: str, unbox: bool):
     """Initializes a variable.
 
     Args:
@@ -366,12 +376,11 @@ class Variable(Generic[T]):
     """Updates the value of this Variable."""
     if self.unbox:
       cur = self.scope.get_variable(self.collection, self.name)
-      cur_struct = tree_util.tree_structure(cur,
-                                            is_leaf=meta.is_axis_metadata)
-      value_struct = tree_util.tree_structure(value,
-                                              is_leaf=meta.is_axis_metadata)
-      has_meta = any(map(meta.is_axis_metadata,
-                         cur_struct.flatten_up_to(cur)))
+      cur_struct = tree_util.tree_structure(cur, is_leaf=meta.is_axis_metadata)
+      value_struct = tree_util.tree_structure(
+        value, is_leaf=meta.is_axis_metadata
+      )
+      has_meta = any(map(meta.is_axis_metadata, cur_struct.flatten_up_to(cur)))
       if cur_struct == value_struct and has_meta:
         value = meta.replace_boxed(cur, value)
 
@@ -384,12 +393,16 @@ class Variable(Generic[T]):
 
 class _ChildRNGSentinel:
   pass
+
+
 # used to identify that an rng counter is meant for a child scope
 child_rng_token = _ChildRNGSentinel()
 
 
 class _DefaultSentinel:
   pass
+
+
 # used to denote no default flag value on scope
 no_flag = _DefaultSentinel()
 
@@ -405,16 +418,20 @@ class Scope:
   <https://github.com/google/flax/tree/main/tests/core/design>`_
   for a number of examples using ``Scopes``.
   """
-  reservations: Dict[str, Optional[str]]
 
-  def __init__(self,
-               variables: MutableVariableDict,
-               rngs: Optional[Dict[str, Union[PRNGKey, LazyRng]]] = None,
-               name: Optional[str] = None,
-               mutable: CollectionFilter = False,
-               parent: Optional['Scope'] = None,
-               path: Iterable[str] = (),
-               flags: Optional[Mapping] = None):
+  reservations: Dict[str, Set[Optional[str]]]
+
+  def __init__(
+    self,
+    variables: MutableVariableDict,
+    rngs: Optional[Union[RNGSequences, Dict[str, LazyRng]]] = None,
+    name: Optional[str] = None,
+    mutable: CollectionFilter = False,
+    parent: Optional['Scope'] = None,
+    path: Iterable[str] = (),
+    debug_path: Iterable[str] = (),
+    flags: Optional[Mapping] = None,
+  ):
     """Initializes a Scope.
 
     Args:
@@ -423,7 +440,9 @@ class Scope:
       name: name of this scope.
       mutable: A CollectionFilter determining which variables are mutable.
       parent: The parent scope.
-      path: The path in the variable tree from the root scope to this scope.
+      path: The path in the variable tree from the root scope to this scope. It
+        exactly matches the module path.
+      debug_path: Similar to path but could contain transformation decorators.
       flags: internal flags.
     """
     rngs = {k: LazyRng.create(v) for k, v in rngs.items()} if rngs else {}
@@ -431,6 +450,7 @@ class Scope:
     self.parent = parent
     self.name = name
     self.path = tuple(path)
+    self.debug_path = tuple(debug_path) or self.path
     self.rngs = rngs
     self.mutable = mutable
     self.flags = freeze({} if flags is None else flags)
@@ -439,7 +459,7 @@ class Scope:
     self.trace_level = tracers.trace_level(tracers.current_trace())
 
     self.rng_counters = {key: 0 for key in self.rngs}
-    self.reservations = dict()
+    self.reservations = collections.defaultdict(set)
 
     self._invalid = False
 
@@ -451,7 +471,11 @@ class Scope:
       return False
     if self is other:
       return True
-    return self.root._variables is other.root._variables and self.path == other.path and self.rng_counters is other.rng_counters
+    return (
+      self.root._variables is other.root._variables
+      and self.path == other.path
+      and self.rng_counters is other.rng_counters
+    )
 
   def __hash__(self) -> int:
     # see __eq__
@@ -463,8 +487,8 @@ class Scope:
 
   @property
   def path_text(self) -> str:
-    """Returns the path as a human readable string with slashes between parts."""
-    return '/' + '/'.join(self.path)
+    """Returns the debug path as a human readable string."""
+    return '/' + '/'.join(self.debug_path)
 
   @property
   def invalid(self) -> bool:
@@ -490,8 +514,9 @@ class Scope:
   def mutable_variables(self) -> Union[VariableDict, Dict[str, Any]]:
     """Returns an immutable copy of the mutable variables belonging to this Scope."""
     self._populate_collections()
-    xs = {k: v for k, v in self._variables.items()
-          if in_filter(self.mutable, k)}
+    xs = {
+      k: v for k, v in self._variables.items() if in_filter(self.mutable, k)
+    }
     if config.flax_return_frozendict:
       return freeze(xs)
     return xs
@@ -517,8 +542,16 @@ class Scope:
       emptied, and the rng counter is optionally rewound.
     """
     self._check_valid()
-    scope = Scope(self._variables, self.rngs, self.name, self.mutable,
-                  self.parent, path=self.path, flags=self.flags)
+    scope = Scope(
+      self._variables,
+      self.rngs,
+      self.name,
+      self.mutable,
+      self.parent,
+      path=self.path,
+      debug_path=self.debug_path,
+      flags=self.flags,
+    )
     if not rewind_rngs:
       scope.rng_counters = self.rng_counters
     return scope
@@ -531,13 +564,13 @@ class Scope:
       col: if a variable, the collection used.
     """
     if name in self.reservations:
-      # with relaxed naming, allow the same name for two variables in
+      # allow the same name for two variables in
       # different collections, otherwise raise error.
-      if config.flax_relaxed_naming:
-        if (self.reservations[name] is None or col is None
-            or self.reservations[name] == col):
-            return True
-      else:
+      if (
+        None in self.reservations[name]
+        or col is None
+        or col in self.reservations[name]
+      ):
         return True
     return False
 
@@ -551,11 +584,12 @@ class Scope:
       col: if a variable, the collection used.
     """
     if not isinstance(name, str):
-      raise TypeError('The type of scope "{name}" should be string but '
-                      f'it is {type(name)}')
+      raise TypeError(
+        'The type of scope "{name}" should be string but ' f'it is {type(name)}'
+      )
     if self.name_reserved(name, col):
       raise ValueError(f'Duplicate use of scope name: "{name}"')
-    self.reservations[name] = col
+    self.reservations[name].add(col)
 
   def default_name(self, prefix: str) -> str:
     """Generates an unreserved name with the given prefix.
@@ -573,10 +607,9 @@ class Scope:
         return name
       i += 1
 
-  def push(self,
-           name: Optional[str] = None,
-           prefix: str = '',
-           reuse=False) -> 'Scope':
+  def push(
+    self, name: Optional[str] = None, prefix: str = '', reuse=False
+  ) -> 'Scope':
     """Creates a child Scope.
 
     Args:
@@ -597,26 +630,31 @@ class Scope:
     rngs = {key: LazyRng.create(rng, name) for key, rng in self.rngs.items()}
     rng_key = (child_rng_token, name)
     if rng_key in self.rng_counters:
-      rng_counters = self.rng_counters.get(rng_key) # type: ignore
+      rng_counters = self.rng_counters.get(rng_key)  # type: ignore
     else:
       rng_counters = {key: 0 for key in rngs}
-      self.rng_counters[rng_key] = rng_counters # type: ignore
-    scope = Scope({},
-                  name=name,
-                  rngs=rngs,
-                  parent=self,
-                  mutable=self.mutable,
-                  path=self.path + (name,),
-                  flags=self.flags)
+      self.rng_counters[rng_key] = rng_counters  # type: ignore
+    scope = Scope(
+      {},
+      name=name,
+      rngs=rngs,
+      parent=self,
+      mutable=self.mutable,
+      path=self.path + (name,),
+      debug_path=self.debug_path + (name,),
+      flags=self.flags,
+    )
     scope.rng_counters = rng_counters
     return scope
 
-  def child(self,
-            fn: Callable[..., Any],
-            name: Optional[str] = None,
-            prefix: Optional[str] = None,
-            named_call: bool = True,
-            **partial_kwargs) -> Callable[..., Any]:
+  def child(
+    self,
+    fn: Callable[..., Any],
+    name: Optional[str] = None,
+    prefix: Optional[str] = None,
+    named_call: bool = True,
+    **partial_kwargs,
+  ) -> Callable[..., Any]:
     """Partially applies a child scope to fn.
 
     When calling the returned function multiple times variables will be reused.
@@ -705,10 +743,13 @@ class Scope:
     """Returns true if a PRNGSequence with name `name` exists."""
     return name in self.rngs
 
-  def make_rng(self, name: str) -> PRNGKey:
+  def make_rng(self, name: str = 'default') -> PRNGKey:
     """Generates A PRNGKey from a PRNGSequence with name `name`."""
     if not self.has_rng(name):
-      raise errors.InvalidRngError(f'{self.name} needs PRNG for "{name}"')
+      if self.has_rng('default'):
+        name = 'default'
+      else:
+        raise errors.InvalidRngError(f'{self.name} needs PRNG for "{name}"')
     self._check_valid()
     self._validate_trace_level()
     self.rng_counters[name] += 1
@@ -760,8 +801,11 @@ class Scope:
     # Make sure reference sharing of child variable dictionaries isn't broken.
     # See https://github.com/google/flax/issues/2022 for more details.
     def put(target, key, val):
-      if (key in target and isinstance(target[key], dict) and
-          isinstance(val, Mapping)):
+      if (
+        key in target
+        and isinstance(target[key], dict)
+        and isinstance(val, Mapping)
+      ):
         for k, v in val.items():
           put(target[key], k, v)
       else:
@@ -769,9 +813,57 @@ class Scope:
 
     put(variables, name, value)
 
-  def variable(self, col: str, name: str,  # pylint: disable=keyword-arg-before-vararg
-               init_fn: Optional[Callable[..., T]] = None,
-               *init_args, unbox: bool = True) -> Variable[T]:
+  @overload
+  def variable(
+    self,
+    col: str,
+    name: str,
+    init_fn: Optional[Callable[..., T]] = None,
+    *init_args,
+  ) -> Variable[T]:
+    ...
+
+  @overload
+  def variable(
+    self,
+    col: str,
+    name: str,
+    init_fn: Optional[Callable[..., T]] = None,
+    *init_args,
+    unbox: Literal[True],
+  ) -> Variable[T]:
+    ...
+
+  @overload
+  def variable(
+    self,
+    col: str,
+    name: str,
+    init_fn: Optional[Callable[..., T]] = None,
+    *init_args,
+    unbox: Literal[False],
+  ) -> Variable[meta.AxisMetadata[T]]:
+    ...
+
+  @overload
+  def variable(
+    self,
+    col: str,
+    name: str,
+    init_fn: Optional[Callable[..., T]] = None,
+    *init_args,
+    unbox: bool = True,
+  ) -> Union[Variable[T], Variable[meta.AxisMetadata[T]]]:
+    ...
+
+  def variable(
+    self,
+    col: str,
+    name: str,  # pylint: disable=keyword-arg-before-vararg
+    init_fn: Optional[Callable[..., T]] = None,
+    *init_args,
+    unbox: bool = True,
+  ) -> Union[Variable[T], Variable[meta.AxisMetadata[T]]]:
     """Creates a variable if it doesn't exist yet in this scope and returns it.
 
     Args:
@@ -795,10 +887,45 @@ class Scope:
         raise errors.ScopeVariableNotFoundError(name, col, self.path_text)
       init_value = init_fn(*init_args)
       self.put_variable(col, name, init_value)
-    return Variable(self, col, name, unbox=unbox)
+    # cast to make static analyzers happy
+    return cast(
+      Union[Variable[T], Variable[meta.AxisMetadata[T]]],
+      Variable(self, col, name, unbox=unbox),
+    )
 
-  def param(self, name: str, init_fn: Callable[..., T], *init_args,
-            unbox: bool = True) -> T:
+  @overload
+  def param(self, name: str, init_fn: Callable[..., T], *init_args) -> T:
+    ...
+
+  @overload
+  def param(
+    self,
+    name: str,
+    init_fn: Callable[..., T],
+    *init_args,
+    unbox: Literal[True],
+  ) -> T:
+    ...
+
+  @overload
+  def param(
+    self,
+    name: str,
+    init_fn: Callable[..., T],
+    *init_args,
+    unbox: Literal[False],
+  ) -> meta.AxisMetadata[T]:
+    ...
+
+  @overload
+  def param(
+    self, name: str, init_fn: Callable[..., T], *init_args, unbox: bool
+  ) -> Union[T, meta.AxisMetadata[T]]:
+    ...
+
+  def param(
+    self, name: str, init_fn: Callable[..., T], *init_args, unbox: bool = True
+  ) -> Union[T, meta.AxisMetadata[T]]:
     """Creates a parameter if it doesn't exist yet in this scope and returns it.
 
     If the parameter exists already, the existing value is simply returned.
@@ -816,15 +943,13 @@ class Scope:
     """
     self.reserve(name, 'params')
     if self.has_variable('params', name):
-      abs_rng = jax.ShapeDtypeStruct(random.default_prng_impl().key_shape,
-                                     jnp.uint32)
       value = self.get_variable('params', name)
       # Validate that the shape of the init_fn output is the same as the shape
       # of the existing parameter. This is to make sure that the hparams set up
       # in a Flax Module match the shapes coming in during apply, and if not,
       # catch it with an error message.
       # NOTE: We could consider moving this to `self.`
-      abs_value = jax.eval_shape(lambda rng: init_fn(rng, *init_args), abs_rng)
+      abs_value = jax.eval_shape(lambda: init_fn(random.key(0), *init_args))
       abs_value_flat = jax.tree_util.tree_leaves(abs_value)
       value_flat = jax.tree_util.tree_leaves(value)
       for val, abs_val in zip(value_flat, abs_value_flat):
@@ -832,8 +957,9 @@ class Scope:
         # usefuleness is less obvious. We might intentionally change the dtype
         # for inference to a half float type for example.
         if jnp.shape(val) != jnp.shape(abs_val):
-          raise errors.ScopeParamShapeError(name, self.path_text,
-                                            jnp.shape(val), jnp.shape(abs_val))
+          raise errors.ScopeParamShapeError(
+            name, self.path_text, jnp.shape(abs_val), jnp.shape(val)
+          )
     else:
       if not self.is_mutable_collection('params'):
         if self.is_collection_empty('params'):
@@ -865,18 +991,20 @@ def _unfreeze_variables(variables, mutable):
     if in_filter(mutable, key):
       new_variables[key] = unfreeze(value)
     else:
-      new_variables[key] = freeze(value)
+      new_variables[key] = value
   return new_variables
 
 
-def bind(variables: VariableDict,
-         rngs: Optional[RNGSequences] = None,
-         mutable: CollectionFilter = False,
-         flags: Optional[Mapping] = None):
+def bind(
+  variables: VariableDict,
+  rngs: Optional[RNGSequences] = None,
+  mutable: CollectionFilter = False,
+  flags: Optional[Mapping] = None,
+):
   """Binds variables and rngs to a new ``Scope``.
 
   bind provides a ``Scope`` instance without transforming a function with
-  ``apply``. This is particalary useful for debugging and interactive use cases
+  ``apply``. This is particularly useful for debugging and interactive use cases
   like notebooks where a function would limit the ability split up code into
   different cells.
 
@@ -888,7 +1016,7 @@ def bind(variables: VariableDict,
   Args:
     variables: Variable dictionary to bind.
     rngs: RNGs to bind.
-    mutable: Which variable colections to treat as mutable.
+    mutable: Which variable collections to treat as mutable.
     flags: internal flags.
 
   Returns:
@@ -898,14 +1026,17 @@ def bind(variables: VariableDict,
     raise errors.ApplyScopeInvalidVariablesTypeError()
   if rngs is not None and not _is_valid_rngs(rngs):
     raise errors.InvalidRngError(
-        'rngs should be a dictionary mapping strings to `jax.PRNGKey`.')
+      'rngs should be a dictionary mapping strings to `jax.PRNGKey`.'
+    )
   new_variables = _unfreeze_variables(variables, mutable)
   return Scope(new_variables, rngs=rngs, mutable=mutable, flags=flags)
 
 
-def apply(fn: Callable[..., Any],
-          mutable: CollectionFilter = False,
-          flags: Optional[Mapping] = None) -> Callable[..., Any]:
+def apply(
+  fn: Callable[..., Any],
+  mutable: CollectionFilter = False,
+  flags: Optional[Mapping] = None,
+) -> Callable[..., Any]:
   """Functionalize a `Scope` function.
 
   Args:
@@ -918,18 +1049,23 @@ def apply(fn: Callable[..., Any],
   """
 
   @functools.wraps(fn)
-  def wrapper(variables: VariableDict,
-              *args,
-              rngs: Optional[RNGSequences] = None,
-              **kwargs) -> Union[Any, Tuple[Any, Union[VariableDict, Dict[str, Any]]]]:
+  def wrapper(
+    variables: VariableDict,
+    *args,
+    rngs: Optional[RNGSequences] = None,
+    **kwargs,
+  ) -> Union[Any, Tuple[Any, Union[VariableDict, Dict[str, Any]]]]:
     # Try to detect if user accidentally passed {'params': {'params': ...}.
-    if 'params' in variables and isinstance(
-        variables['params'],
-        (dict, FrozenDict)) and 'params' in variables['params']:
+    if (
+      'params' in variables
+      and isinstance(variables['params'], (dict, FrozenDict))
+      and 'params' in variables['params']
+    ):
       raise errors.ApplyScopeInvalidVariablesStructureError(variables)
 
-    with bind(variables, rngs=rngs, mutable=mutable,
-              flags=flags).temporary() as root:
+    with bind(
+      variables, rngs=rngs, mutable=mutable, flags=flags
+    ).temporary() as root:
       y = fn(root, *args, **kwargs)
     if mutable is not False:
       return y, root.mutable_variables()
@@ -939,9 +1075,11 @@ def apply(fn: Callable[..., Any],
   return wrapper
 
 
-def init(fn: Callable[..., Any],
-         mutable: CollectionFilter = True,
-         flags: Optional[Mapping] = None) -> Callable[..., Any]:
+def init(
+  fn: Callable[..., Any],
+  mutable: CollectionFilter = True,
+  flags: Optional[Mapping] = None,
+) -> Callable[..., Any]:
   """Functionalize a `Scope` function for initialization.
 
   Args:
@@ -956,21 +1094,26 @@ def init(fn: Callable[..., Any],
   @functools.wraps(fn)
   def wrapper(rngs, *args, **kwargs) -> Tuple[Any, VariableDict]:
     if not _is_valid_rng(rngs) and not _is_valid_rngs(rngs):
-      raise ValueError('First argument passed to an init function should be a '
-                       '`jax.PRNGKey` or a dictionary mapping strings to '
-                       '`jax.PRNGKey`.')
+      raise ValueError(
+        'First argument passed to an init function should be a '
+        '`jax.PRNGKey` or a dictionary mapping strings to '
+        '`jax.PRNGKey`.'
+      )
     if not isinstance(rngs, dict):
       rngs = {'params': rngs}
     init_flags = {**(flags if flags is not None else {}), 'initializing': True}
-    return apply(fn, mutable=mutable, flags=init_flags)({}, *args, rngs=rngs,
-                                                        **kwargs)
+    return apply(fn, mutable=mutable, flags=init_flags)(
+      {}, *args, rngs=rngs, **kwargs
+    )
 
   return wrapper
 
 
-def lazy_init(fn: Callable[..., Any],
-         mutable: CollectionFilter = True,
-         flags: Optional[Mapping] = None) -> Callable[..., Any]:
+def lazy_init(
+  fn: Callable[..., Any],
+  mutable: CollectionFilter = True,
+  flags: Optional[Mapping] = None,
+) -> Callable[..., Any]:
   """Functionalizes a `Scope` function for lazy initialization.
 
   Similair to ``init`` except that the init function now accepts
@@ -985,7 +1128,7 @@ def lazy_init(fn: Callable[..., Any],
         k = scope.param("kernel", nn.initializers.lecun_normal(), (x.shape[-1], x.shape[-1]))
         return x @ k
     init_fn = lazy_init(f)
-    variables = init_fn(random.PRNGKey(0), jax.ShapeDtypeStruct((1, 128), jnp.float32))
+    variables = init_fn(random.key(0), jax.ShapeDtypeStruct((1, 128), jnp.float32))
 
 
   Args:
@@ -997,7 +1140,9 @@ def lazy_init(fn: Callable[..., Any],
     `fn` with the scope partially applied. Unlike ``init`` which returns a tuple of function
     output and variables, the lazy init function only returns the variables.
   """
-  return partial_eval.lazy_init(lambda *args, **kwargs: init(fn, mutable, flags)(*args, **kwargs)[1])
+  return partial_eval.lazy_init(
+    lambda *args, **kwargs: init(fn, mutable, flags)(*args, **kwargs)[1]
+  )
 
 
 def _is_valid_collection(col: VariableDict):
@@ -1029,17 +1174,24 @@ def _is_valid_variables(variables: VariableDict) -> bool:
 
 def _is_valid_rng(rng: Array):
   """Checks whether rng is a valid JAX PRNGKey, also handling custom prngs."""
-  # New-style JAX KeyArrays have a base type.
-  if jax_config.jax_enable_custom_prng: # type: ignore[attr-defined]
-    if not isinstance(rng, jax.random.KeyArray):
-      return False
-  # Old-style JAX PRNGKeys are plain uint32 arrays.
-  else:
-    if not isinstance(rng, (np.ndarray, jnp.ndarray)):
-      return False
-    if (rng.shape != random.default_prng_impl().key_shape or
-        rng.dtype != jnp.uint32):
-      return False
+  # This check is valid for either new-style or old-style PRNG keys
+  if not isinstance(rng, (np.ndarray, jnp.ndarray)):
+    return False
+
+  # Handle new-style typed PRNG keys
+  if hasattr(jax.dtypes, 'prng_key'):  # JAX 0.4.14 or newer
+    if jax.dtypes.issubdtype(rng.dtype, jax.dtypes.prng_key):
+      return rng.shape == ()
+  elif hasattr(jax.random, 'PRNGKeyArray'):  # Previous JAX versions
+    if isinstance(rng, jax.random.PRNGKeyArray):
+      return rng.shape == ()
+
+  # Handle old-style raw PRNG keys
+  expected_rng = jax.eval_shape(
+    lambda s: jax.random.key_data(jax.random.key(s)), 0
+  )
+  if (rng.shape, rng.dtype) != (expected_rng.shape, expected_rng.dtype):
+    return False
   return True
 
 
