@@ -1,4 +1,4 @@
-# Copyright 2022 The Flax Authors.
+# Copyright 2024 The Flax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -50,9 +50,10 @@ class Template(FlaxError):
 
 
 class FlaxError(Exception):
-
   def __init__(self, message):
-    error_page = 'https://flax.readthedocs.io/en/latest/api_reference/flax.errors.html'
+    error_page = (
+      'https://flax.readthedocs.io/en/latest/api_reference/flax.errors.html'
+    )
     module_name = self.__class__.__module__
     class_name = self.__class__.__name__
     error_msg = f'{message} ({error_page}#{module_name}.{class_name})'
@@ -81,13 +82,14 @@ class LazyInitError(FlaxError):
         # this causes an error when using lazy_init.
         k = self.param("kernel", lambda _: x)
         return x * k
-    Foo().lazy_init(random.PRNGKey(0), jax.ShapeDtypeStruct((8, 4), jnp.float32))
+    Foo().lazy_init(random.key(0), jax.ShapeDtypeStruct((8, 4), jnp.float32))
   """
 
   def __init__(self, partial_val):
     super().__init__(
-        f'Lazy init encoutered a value that could with '
-        f'the given inputs (shape: {partial_val}).')
+      'Lazy init encountered a value that could with '
+      f'the given inputs (shape: {partial_val}).'
+    )
 
 
 #################################################
@@ -128,7 +130,7 @@ class InvalidRngError(FlaxError):
 
   So, ``Foo`` is initialized as follows::
 
-    init_rngs = {'params': random.PRNGKey(0), 'dropout': random.PRNGKey(1)}
+    init_rngs = {'params': random.key(0), 'dropout': random.key(1)}
     variables = Foo().init(init_rngs, init_inputs)
 
   If a Module only requires an rng for ``params``, you can use::
@@ -141,7 +143,7 @@ class InvalidRngError(FlaxError):
   When applying ``Foo``, only the rng for ``dropout`` is needed, because
   ``params`` is only used for initializing the Module parameters::
 
-    Foo().apply(variables, inputs, rngs={'dropout': random.PRNGKey(2)})
+    Foo().apply(variables, inputs, rngs={'dropout': random.key(2)})
 
   If a Module only requires an rng for ``params``, you don't have to provide
   rngs for apply at all::
@@ -164,13 +166,15 @@ class ApplyScopeInvalidVariablesTypeError(FlaxError):
   """
 
   def __init__(self):
-    super().__init__('The first argument passed to an apply function should be '
-                     'a dictionary of collections. Each collection should be a '
-                     'dictionary with string keys.')
+    super().__init__(
+      'The first argument passed to an apply function should be '
+      'a dictionary of collections. Each collection should be a '
+      'dictionary with string keys.'
+    )
 
 
 class ApplyScopeInvalidVariablesStructureError(FlaxError):
-  """This error is thrown when the dict passed as `variables` to apply() has an
+  """This error is thrown when the dict passed as ``variables`` to apply() has an
 
   extra 'params' layer, i.e. {'params': {'params': ...}}.
   For more explanation on variable dicts, please see :mod:`flax.core.variables`.
@@ -178,10 +182,11 @@ class ApplyScopeInvalidVariablesStructureError(FlaxError):
 
   def __init__(self, variables):
     super().__init__(
-        'Expect the `variables` (first argument) passed to apply() '
-        'to be a dict with the structure {"params": ...}, but got a dict '
-        'with an extra params layer, i.e.  {"params": {"params": ... } }. '
-        f'You should instead pass in your dict\'s ["params"].')
+      f'Expect the `variables` (first argument) passed to apply() '
+      f'to be a dict with the structure {{"params": ...}}, but got a dict '
+      f'with an extra params layer, i.e.  {{"params": {{"params": ... }} }}. '
+      f'You should instead pass in your dict\'s ["params"].'
+    )
 
 
 class ScopeParamNotFoundError(FlaxError):
@@ -203,13 +208,15 @@ class ScopeParamNotFoundError(FlaxError):
         return embedding[inputs]
 
     model = Embed(4, 8)
-    variables = model.init(random.PRNGKey(0), jnp.ones((5, 5, 1)))
+    variables = model.init(random.key(0), jnp.ones((5, 5, 1)))
     _ = model.apply(variables, jnp.ones((5, 5, 1)), 'embed')
   """
 
   def __init__(self, param_name, scope_path):
-    super().__init__(f'Could not find parameter named "{param_name}" in scope '
-                     f'"{scope_path}".')
+    super().__init__(
+      f'Could not find parameter named "{param_name}" in scope '
+      f'"{scope_path}".'
+    )
 
 
 class ScopeCollectionNotFound(FlaxError):
@@ -217,6 +224,7 @@ class ScopeCollectionNotFound(FlaxError):
   collection.
 
   There are two common causes:
+
   1. | The collection was not passed to ``apply`` correctly.
      | For example, you might have used ``module.apply(params, ...)`` instead
      | of ``module.apply({'params': params}, ...)``.
@@ -227,8 +235,9 @@ class ScopeCollectionNotFound(FlaxError):
 
   def __init__(self, col_name, var_name, scope_path):
     super().__init__(
-        f'Tried to access "{var_name}" from collection "{col_name}" in '
-        f'"{scope_path}" but the collection is empty.')
+      f'Tried to access "{var_name}" from collection "{col_name}" in '
+      f'"{scope_path}" but the collection is empty.'
+    )
 
 
 class ScopeParamShapeError(FlaxError):
@@ -255,14 +264,16 @@ class ScopeParamShapeError(FlaxError):
                             (((x.ndim - 1,), (0,)), ((), ())))
         return y
 
-    variables = NoBiasDense().init(random.PRNGKey(0), jnp.ones((5, 5, 1)))
+    variables = NoBiasDense().init(random.key(0), jnp.ones((5, 5, 1)))
     _ = NoBiasDense().apply(variables, jnp.ones((5, 5)))
   """
 
   def __init__(self, param_name, scope_path, value_shape, init_shape):
-    super().__init__('Inconsistent shapes between value and initializer '
-                     f'for parameter "{param_name}" in "{scope_path}": '
-                     f'{value_shape}, {init_shape}.')
+    super().__init__(
+      f'Initializer expected to generate shape {init_shape} '
+      f'but got shape {value_shape} instead for parameter '
+      f'"{param_name}" in "{scope_path}".'
+    )
 
 
 class ScopeVariableNotFoundError(FlaxError):
@@ -274,8 +285,10 @@ class ScopeVariableNotFoundError(FlaxError):
   """
 
   def __init__(self, name, col, scope_path):
-    super().__init__(f'No Variable named "{name}" for collection "{col}" '
-                     f'exists in "{scope_path}".')
+    super().__init__(
+      f'No Variable named "{name}" for collection "{col}" '
+      f'exists in "{scope_path}".'
+    )
 
 
 class InvalidFilterError(FlaxError):
@@ -288,10 +301,10 @@ class InvalidFilterError(FlaxError):
 class InvalidScopeError(FlaxError):
   """A temporary Scope is only valid within the context in which it is created::
 
-    with Scope(variables, rngs=rngs).temporary() as root:
-      y = fn(root, *args, **kwargs)
-      # Here root is valid.
-    # Here root is invalid.
+  with Scope(variables, rngs=rngs).temporary() as root:
+    y = fn(root, *args, **kwargs)
+    # Here root is valid.
+  # Here root is invalid.
   """
 
   def __init__(self, scope_name):
@@ -319,8 +332,10 @@ class ModifyScopeVariableError(FlaxError):
   """
 
   def __init__(self, col, variable_name, scope_path):
-    super().__init__(f'Cannot update variable "{variable_name}" in '
-                     f'"{scope_path}" because collection "{col}" is immutable.')
+    super().__init__(
+      f'Cannot update variable "{variable_name}" in '
+      f'"{scope_path}" because collection "{col}" is immutable.'
+    )
 
 
 class JaxTransformError(FlaxError):
@@ -350,8 +365,9 @@ class PartitioningUnspecifiedError(FlaxError):
 
   def __init__(self, target):
     super().__init__(
-        f'Trying to transform a Partitioned variable but "partition_name"'
-        f' is not specified in metadata_params: {target}')
+      'Trying to transform a Partitioned variable but "partition_name"'
+      f' is not specified in metadata_params: {target}'
+    )
 
 
 #################################################
@@ -410,8 +426,10 @@ class NameInUseError(FlaxError):
 
   def __init__(self, key_type, value, module_name):
     # key_type is in {param, variable, submodule}.
-    super().__init__(f'Could not create {key_type} "{value}" in Module '
-                     f'{module_name}: Name in use.')
+    super().__init__(
+      f'Could not create {key_type} "{value}" in Module '
+      f'{module_name}: Name in use.'
+    )
 
 
 class AssignSubModuleError(FlaxError):
@@ -432,7 +450,7 @@ class AssignSubModuleError(FlaxError):
       def __call__(self, x):
         conv = nn.Conv(features=3, kernel_size=3)
 
-    Foo().init(random.PRNGKey(0), jnp.zeros((1,)))
+    Foo().init(random.key(0), jnp.zeros((1,)))
 
   Note that this error is also thrown if you partially defined a Module inside
   setup::
@@ -445,7 +463,7 @@ class AssignSubModuleError(FlaxError):
         x = self.conv(kernel_size=4)(x)
         return x
 
-    Foo().init(random.PRNGKey(0), jnp.zeros((1,)))
+    Foo().init(random.key(0), jnp.zeros((1,)))
 
   In this case, ``self.conv(kernel_size=4)`` is called from ``__call__``, which
   is disallowed because it's neither within ``setup`` nor a method wrapped in
@@ -453,8 +471,10 @@ class AssignSubModuleError(FlaxError):
   """
 
   def __init__(self, cls):
-    super().__init__(f'Submodule {cls} must be defined in `setup()` or in a '
-                     'method wrapped in `@compact`')
+    super().__init__(
+      f'Submodule {cls} must be defined in `setup()` or in a '
+      'method wrapped in `@compact`'
+    )
 
 
 class SetAttributeInModuleSetupError(FlaxError):
@@ -471,7 +491,7 @@ class SetAttributeInModuleSetupError(FlaxError):
       def __call__(self, x):
         return nn.Dense(self.features)(x)
 
-    variables = SomeModule().init(random.PRNGKey(0), jnp.ones((1, )))
+    variables = SomeModule().init(random.key(0), jnp.ones((1, )))
 
   Instead, these attributes should be set when initializing the Module::
 
@@ -482,7 +502,7 @@ class SetAttributeInModuleSetupError(FlaxError):
       def __call__(self, x):
         return nn.Dense(self.features)(x)
 
-    variables = SomeModule(features=3).init(random.PRNGKey(0), jnp.ones((1, )))
+    variables = SomeModule(features=3).init(random.key(0), jnp.ones((1, )))
 
   TODO(marcvanzee): Link to a design note explaining why it's necessary for
   modules to stay frozen (otherwise we can't safely clone them, which we use for
@@ -509,7 +529,7 @@ class SetAttributeFrozenModuleError(FlaxError):
         x = nn.Dense(self.num_features)(x)
         return x
 
-    s = SomeModule().init(random.PRNGKey(0), jnp.ones((5, 5)))
+    s = SomeModule().init(random.key(0), jnp.ones((5, 5)))
 
   Similarly, the error is raised when trying to modify a submodule's attributes
   after constructing it, even if this is done in the ``setup()`` method of the
@@ -525,9 +545,11 @@ class SetAttributeFrozenModuleError(FlaxError):
   """
 
   def __init__(self, module_cls, attr_name, attr_val):
-    super().__init__(f'Can\'t set {attr_name}={attr_val} for Module of type '
-                     f'{module_cls}: Module instance is frozen outside of '
-                     'setup method.')
+    super().__init__(
+      f"Can't set {attr_name}={attr_val} for Module of type "
+      f'{module_cls}: Module instance is frozen outside of '
+      'setup method.'
+    )
 
 
 class MultipleMethodsCompactError(FlaxError):
@@ -540,7 +562,7 @@ class MultipleMethodsCompactError(FlaxError):
   * Use two separate modules that both have a unique ``@compact`` method.
 
   TODO(marcvanzee): Link to a design note explaining the motivation behind this.
-  There is no need for an equivalent to `hk.transparent` and it makes submodules
+  There is no need for an equivalent to ``hk.transparent`` and it makes submodules
   much more sane because there is no need to prefix the method names.
   """
 
@@ -558,8 +580,9 @@ class ReservedModuleAttributeError(FlaxError):
   """
 
   def __init__(self, annotations):
-    super().__init__(f'properties `parent` and `name` are reserved: '
-                     f'{annotations}')
+    super().__init__(
+      f'properties `parent` and `name` are reserved: {annotations}'
+    )
 
 
 class ApplyModuleInvalidMethodError(FlaxError):
@@ -574,8 +597,9 @@ class ApplyModuleInvalidMethodError(FlaxError):
   """
 
   def __init__(self, method):
-    super().__init__(f'Cannot call apply(): {method} is not a valid function '
-                     'for apply().')
+    super().__init__(
+      f'Cannot call apply(): {method} is not a valid function for apply().'
+    )
 
 
 class CallCompactUnboundModuleError(FlaxError):
@@ -595,17 +619,17 @@ class CallCompactUnboundModuleError(FlaxError):
   :meth:`Module.init() <flax.linen.Module.init>` to get initial variables)::
 
     from jax import random
-    variables = test_dense.init(random.PRNGKey(0), jnp.ones((5,5)))
+    variables = test_dense.init(random.key(0), jnp.ones((5,5)))
 
     y = test_dense.apply(variables, jnp.ones((5,5)))
   """
 
   def __init__(self):
-    super().__init__('Can\'t call compact methods on unbound modules')
+    super().__init__("Can't call compact methods on unbound modules")
 
 
 class CallSetupUnboundModuleError(FlaxError):
-  """This error occurs when you are trying to call `.setup()` directly.
+  """This error occurs when you are trying to call ``.setup()`` directly.
 
   For instance, the error will be raised when trying to run this code::
 
@@ -620,12 +644,12 @@ class CallSetupUnboundModuleError(FlaxError):
     module.setup() # <-- ERROR!
     submodule = module.submodule
 
-  In general you shouldn't call `.setup()` yourself, if you need to get access
-  to a field or submodule defined inside `setup` you can instead create a
+  In general you shouldn't call ``.setup()`` yourself, if you need to get access
+  to a field or submodule defined inside ``setup`` you can instead create a
   function
-  to extract it and pass it to `nn.apply`::
+  to extract it and pass it to ``nn.apply``::
 
-    # setup() will be called automatically by `nn.apply`
+    # setup() will be called automatically by ``nn.apply``
     def get_submodule(module):
       return module.submodule.clone() # avoid leaking the Scope
 
@@ -634,7 +658,8 @@ class CallSetupUnboundModuleError(FlaxError):
   """
 
   def __init__(self):
-    super().__init__('Can\'t call compact methods on unbound modules')
+    super().__init__("Can't call compact methods on unbound modules")
+
 
 class CallUnbindOnUnboundModuleError(FlaxError):
   """This error occurs when you are trying to call ``.unbind()`` on an unbound
@@ -658,11 +683,13 @@ class CallUnbindOnUnboundModuleError(FlaxError):
     ... # do something with bound_module
     module = bound_module.unbind() # <-- OK!
   """
+
   def __init__(self):
-    super().__init__('Can\'t call `unbind()` on unbound modules')
+    super().__init__("Can't call `unbind()` on unbound modules")
+
 
 class InvalidInstanceModuleError(FlaxError):
-  """This error occurs when you are trying to call `.init()`, `.init_with_output()`, `.apply() or `.bind()`
+  """This error occurs when you are trying to call ``.init()``, ``.init_with_output()``, ``.apply()`` or ``.bind()``
 
   on the Module class itself, instead of an instance of the Module class.
   For example, the error will be raised when trying to run this code::
@@ -672,8 +699,8 @@ class InvalidInstanceModuleError(FlaxError):
       def __call__(self, x):
         return x
 
-    k = random.PRNGKey(0)
-    x = random.uniform(random.PRNGKey(1), (2,))
+    k = random.key(0)
+    x = random.uniform(random.key(1), (2,))
     B.init(k, x)   # B is module class, not B() a module instance
     B.apply(vs, x)   # similar issue with apply called on class instead of
     instance.
@@ -681,12 +708,13 @@ class InvalidInstanceModuleError(FlaxError):
 
   def __init__(self):
     super().__init__(
-        'Can only call init, init_with_output or apply methods on an instance of the Module class, not the Module class itself'
+      'Can only call init, init_with_output or apply methods on an instance'
+      ' of the Module class, not the Module class itself'
     )
 
 
 class IncorrectPostInitOverrideError(FlaxError):
-  """This error occurs when you overrode `.__post_init__()` without calling `super().__post_init__()`.
+  """This error occurs when you overrode ``.__post_init__()`` without calling ``super().__post_init__()``.
 
   For example, the error will be raised when trying to run this code::
 
@@ -703,12 +731,13 @@ class IncorrectPostInitOverrideError(FlaxError):
         return input + 3
 
     r = A(x=3)
-    r.init(jax.random.PRNGKey(2), jnp.ones(3))
+    r.init(jax.random.key(2), jnp.ones(3))
   """
 
   def __init__(self):
     super().__init__(
-        'Overrode `.__post_init__()` without calling `super().__post_init__()`')
+      'Overrode `.__post_init__()` without calling `super().__post_init__()`'
+    )
 
 
 class DescriptorAttributeError(FlaxError):
@@ -724,12 +753,13 @@ class DescriptorAttributeError(FlaxError):
             return self.prop
 
     foo = Foo()
-    variables = foo.init(jax.random.PRNGKey(0), jnp.ones(shape=(1, 8)))
+    variables = foo.init(jax.random.key(0), jnp.ones(shape=(1, 8)))
   """
 
   def __init__(self):
     super().__init__(
-        'Trying to access a property that is accessing a non-existent attribute.'
+      'Trying to access a property that is accessing a non-existent'
+      ' attribute.'
     )
 
 
@@ -744,7 +774,8 @@ class InvalidCheckpointError(FlaxError):
 
   def __init__(self, path, step):
     super().__init__(
-        f'Trying to save an outdated checkpoint at step: "{step}" and path: "{path}".'
+      f'Trying to save an outdated checkpoint at step: "{step}" and path:'
+      f' "{path}".'
     )
 
 
@@ -761,9 +792,10 @@ class MPACheckpointingRequiredError(FlaxError):
 
   def __init__(self, path, step):
     super().__init__(
-        f'Checkpoint failed at step: "{step}" and path: "{path}": Target '
-        'contains a multiprocess array should be saved/restored with a '
-        'GlobalAsyncCheckpointManager.')
+      f'Checkpoint failed at step: "{step}" and path: "{path}": Target '
+      'contains a multiprocess array should be saved/restored with a '
+      'GlobalAsyncCheckpointManager.'
+    )
 
 
 class MPARestoreTargetRequiredError(FlaxError):
@@ -778,10 +810,11 @@ class MPARestoreTargetRequiredError(FlaxError):
 
   def __init__(self, path, step, key=None):
     error_msg = (
-        f'Restore checkpoint failed at step: "{step}" and path: "{path}": '
-        'Checkpoints containing a multiprocess array need to be restored with '
-        'a target with pre-created arrays. If you cannot provide a full valid '
-        'target, consider ``allow_partial_mpa_restoration=True``. ')
+      f'Restore checkpoint failed at step: "{step}" and path: "{path}": '
+      'Checkpoints containing a multiprocess array need to be restored with '
+      'a target with pre-created arrays. If you cannot provide a full valid '
+      'target, consider ``allow_partial_mpa_restoration=True``. '
+    )
     if key:
       error_msg += f'This error fired when trying to restore array at {key}.'
     super().__init__(error_msg)
@@ -795,25 +828,10 @@ class MPARestoreDataCorruptedError(FlaxError):
 
   def __init__(self, step, path):
     super().__init__(
-        f'Restore checkpoint failed at step: "{step}" on multiprocess array at '
-        f' "{path}": No "commit_success.txt" found on this "_gda" directory. '
-        'Was its save halted before completion?')
-
-
-class MPARestoreTypeNotMatchError(FlaxError):
-  """Make sure the multiprocess array type you use matches your configuration in jax.config.jax_array.
-
-  If you turned `jax.config.jax_array` on, you should use
-  `jax.experimental.array.Array` everywhere, instead of using
-  `GlobalDeviceArray`. Otherwise, avoid using jax.experimental.array
-  to restore your checkpoint.
-  """
-
-  def __init__(self, step, gda_path):
-    super().__init__(
-        f'Restore checkpoint failed at step: "{step}" on multiprocess array at '
-        f' "{gda_path}": The array type provided by the target does not match '
-        'the JAX global configuration, namely the jax.config.jax_array.')
+      f'Restore checkpoint failed at step: "{step}" on multiprocess array at '
+      f' "{path}": No "commit_success.txt" found on this "_gda" directory. '
+      'Was its save halted before completion?'
+    )
 
 
 #################################################
@@ -826,7 +844,8 @@ class TransformedMethodReturnValueError(FlaxError):
 
   def __init__(self, name):
     super().__init__(
-        f'Transformed module method {name} cannot return Modules or Variables.')
+      f'Transformed module method {name} cannot return Modules or Variables.'
+    )
 
 
 class TransformTargetError(FlaxError):
@@ -838,7 +857,7 @@ class TransformTargetError(FlaxError):
 
     nn.vmap(nn.Dense(features))(x)  # raises TransformTargetError
 
-  You can transform the `nn.Dense` class directly instead::
+  You can transform the ``nn.Dense`` class directly instead::
 
     nn.vmap(nn.Dense)(features)(x)
 
@@ -856,9 +875,10 @@ class TransformTargetError(FlaxError):
 
   def __init__(self, target):
     super().__init__(
-        'Linen transformations must be applied to Modules classes or'
-        ' functions taking a Module instance as the first argument.'
-        f' The provided target is not a Module class or callable: {target}')
+      'Linen transformations must be applied to Modules classes or'
+      ' functions taking a Module instance as the first argument.'
+      f' The provided target is not a Module class or callable: {target}'
+    )
 
 
 #################################################
@@ -875,3 +895,53 @@ class AlreadyExistsError(FlaxError):
 
   def __init__(self, path):
     super().__init__(f'Trying overwrite an existing file: "{path}".')
+
+
+#################################################
+# cursor.py errors                                  #
+#################################################
+
+
+class CursorFindError(FlaxError):
+  """Error when calling :meth:`Cursor.find() <flax.cursor.Cursor.find>`.
+
+  This error occurs if no object or more than one object is found, given
+  the conditions of the ``cond_fn``.
+  """
+
+  def __init__(self, cursor=None, cursor2=None):
+    if cursor and cursor2:
+      super().__init__(
+        'More than one object found given the conditions of the cond_fn. '
+        'The first two objects found have the following paths: '
+        f'{cursor._path} and {cursor2._path}'
+      )
+    else:
+      super().__init__('No object found given the conditions of the cond_fn.')
+
+
+class TraverseTreeError(FlaxError):
+  """Error when calling ``Cursor._traverse_tree()``. This function has two
+  modes:
+
+  - if ``update_fn`` is not None, it will traverse the tree and return a
+    generator of tuples containing the path where the ``update_fn`` was
+    applied and the newly modified value.
+  - if ``cond_fn`` is not None, it will traverse the tree and return a
+    generator of tuple paths that fulfilled the conditions of the ``cond_fn``.
+
+  This error occurs if either both ``update_fn`` and ``cond_fn`` are None,
+  or both are not None.
+  """
+
+  def __init__(self, update_fn, cond_fn):
+    if update_fn is None and cond_fn is None:
+      super().__init__(
+        'Both update_fn and cond_fn are None. Exactly one of them must be'
+        ' None.'
+      )
+    else:
+      super().__init__(
+        'Both update_fn and cond_fn are not None. Exactly one of them must be'
+        ' not None.'
+      )
