@@ -1,4 +1,4 @@
-# Copyright 2022 The Flax Authors.
+# Copyright 2024 The Flax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,20 +12,19 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-"""Write Summaries from JAX for use with Tensorboard.
-"""
+"""Write Summaries from JAX for use with Tensorboard."""
 
 import contextlib
 import functools
 import os
 
-# pylint: disable=g-import-not-at-top
-from flax import io
 import numpy as np
-
 import tensorflow as tf  # pytype: disable=import-error
 from tensorboard.plugins.hparams import api as hparams_api
 
+# pylint: disable=g-import-not-at-top
+from flax import io
+
 
 def _flatten_dict(input_dict, parent_key='', sep='.'):
   """Flattens and simplifies dict such that it can be used by hparams.
@@ -43,7 +42,16 @@ def _flatten_dict(input_dict, parent_key='', sep='.'):
     new_key = parent_key + sep + k if parent_key else k
 
     # Valid types according to https://github.com/tensorflow/tensorboard/blob/1204566da5437af55109f7a4af18f9f8b7c4f864/tensorboard/plugins/hparams/summary_v2.py
-    valid_types = (bool, int, float, str, np.bool_, np.integer, np.floating, np.character)
+    valid_types = (
+      bool,
+      int,
+      float,
+      str,
+      np.bool_,
+      np.integer,
+      np.floating,
+      np.character,
+    )
 
     if isinstance(v, dict):
       # Recursively flatten the dict.
@@ -167,8 +175,13 @@ class SummaryWriter:
     audio = tf.convert_to_tensor(audiodata, dtype=tf.float32)
     with self._as_default(self._event_writer):
       tf.summary.audio(
-          name=tag, data=audio, sample_rate=sample_rate, step=step,
-          max_outputs=max_outputs, encoding='wav')
+        name=tag,
+        data=audio,
+        sample_rate=sample_rate,
+        step=step,
+        max_outputs=max_outputs,
+        encoding='wav',
+      )
 
   def histogram(self, tag, values, step, bins=None):
     """Saves histogram of values.
@@ -199,7 +212,7 @@ class SummaryWriter:
       tf.summary.text(name=tag, data=tf.constant(textdata), step=step)
 
   def write(self, tag, tensor, step, metadata=None):
-    """Saves a arbitrary tensor summary.
+    """Saves an arbitrary tensor summary.
 
     Useful when working with custom plugins or constructing a summary directly.
 
@@ -211,11 +224,7 @@ class SummaryWriter:
     Note: markdown formatting is rendered by tensorboard.
     """
     with self._as_default(self._event_writer):
-      tf.summary.write(
-          tag=tag,
-          tensor=tensor,
-          step=step,
-          metadata=metadata)
+      tf.summary.write(tag=tag, tensor=tensor, step=step, metadata=metadata)
 
   def hparams(self, hparams):
     """Saves hyper parameters.
